use std::net::SocketAddr;

use axum::routing::get;
use axum::Router;
use clap::{Parser, Subcommand};
use grpc::{greeter_server, HelloReply, HelloRequest};
use tokio::net::TcpListener;
use tonic::transport::Server;

#[derive(Parser)]
#[command(author, version, about, long_about = None, subcommand_required = true)]
struct Command {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    Serve {
        #[arg(env = "SERVICE_HOST", long, default_value = "127.0.0.1:3000")]
        host: SocketAddr,
    },
}

async fn axum_serve(listener: TcpListener, app: Router) -> anyhow::Result<()> {
    axum::serve(listener, app.into_make_service()).await?;

    Ok(())
}

mod grpc {
    tonic::include_proto!("flux_releaser");
}

#[derive(Debug, Default)]
pub struct FluxReleaserGrpc {}
#[tonic::async_trait]
impl greeter_server::Greeter for FluxReleaserGrpc {
    async fn say_hello(
        &self,
        request: tonic::Request<HelloRequest>,
    ) -> std::result::Result<tonic::Response<HelloReply>, tonic::Status> {
        Ok(tonic::Response::new(HelloReply {
            message: "something".into(),
        }))
    }
}

async fn tonic_serve() -> anyhow::Result<()> {
    tracing::info!("grpc listening on: :7900");
    Server::builder()
        .add_service(greeter_server::GreeterServer::new(
            FluxReleaserGrpc::default(),
        ))
        .serve("[::1]:7900".parse()?)
        .await?;

    Ok(())
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenv::dotenv().ok();
    tracing_subscriber::fmt::init();

    let cli = Command::parse();

    if let Some(Commands::Serve { host }) = cli.command {
        tracing::info!("Starting service");

        let app = Router::new().route("/", get(root));

        tracing::info!("listening on {}", host);
        let listener = tokio::net::TcpListener::bind(host).await.unwrap();

        tokio::select! {
            res = axum_serve(listener, app) => {
              res?;
            },
            res = tonic_serve() => {
              res?;
            },
        };
    }

    Ok(())
}

async fn root() -> &'static str {
    "Hello, flux-releaser!"
}
